/* DChub - a Direct Connect Hub for Linux
 *
 * Copyright (C) 2001 Eric Prevoteau
 *
 * xmltools.c: Copyright (C) Yves BLUSSEAU
 *
 * $Id: users_xml.c,v 2.2 2003/09/28 12:16:25 blusseau Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <string.h>
#include <glib.h>
#include <sys/stat.h>
#include <time.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include "hl_locks.h"
#include "xmltools.h"
#include "users_xml.h"

#ifdef HAVE_LIBDMALLOCTH
#include <dmalloc.h>    /* Gray Watson's library */
#define show_alloc() dmalloc_log_unfreed()
#define show_stats() dmalloc_log_stats()
#else
#define show_alloc()    /* nothing */
#define show_stats()    /* nothing */
#endif

static char *xml_file="users.xml";
time_t time_user_file_loaded=0;
G_LOCK_DEFINE(users_doc_lock);
xmlDocPtr user_doc=NULL;
HL_MUTEX  levels_table_lock=HL_MUTEX_INIT;
GHashTable *levels_hash_table=NULL;

void refresh_level_table() {
	xmlXPathObjectPtr search;
	HL_LOCK_WRITE(levels_table_lock);
	if (levels_hash_table) {
		/* Destroy the old hash */
		g_hash_table_destroy(levels_hash_table);
	}
	levels_hash_table=g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	search = getNodeSet (user_doc,"//privileges/access"); /* retrieve all access */
	if (search) {
		int i;
		xmlNodeSetPtr nodeset = search->nodesetval;
		for (i=0; i < nodeset->nodeNr; i++) {
			xmlNodePtr xml_access_node=nodeset->nodeTab[i];
			xmlChar *xml_name=xmlGetProp(xml_access_node, "name"); /* get the access node name */
			if (xml_name) {
				xmlChar *xml_level=xmlGetProp(xml_access_node, "level"); /* get the level of the node */
				if (xml_level) {
					gint* value=g_new(int,1);
					*value=xmlXPathCastStringToNumber(xml_level);
					// printf("Access Name:[%s] Level:[%d]\n",xml_name,*value);
					g_hash_table_insert(levels_hash_table, g_strdup(xml_name), value);
					xmlFree(xml_level);
				}
				xmlFree(xml_name);
			}
		}
		xmlXPathFreeObject(search);
	}
	HL_UNLOCK_WRITE(levels_table_lock);
}

gboolean refresh_user_doc() {
	struct stat user_file_stat;
	if (stat(xml_file, &user_file_stat)) {
		return user_doc ? TRUE : FALSE; /* ok if a DOM already in memory */
	}
	if (time_user_file_loaded < user_file_stat.st_atime) {
		/* load the new file in memory */
		xmlDocPtr new_user_doc=loadXMLFile(xml_file);
		if (new_user_doc) {			
			/* replace old DOM with the new one */
			G_LOCK(users_doc_lock);
			xmlFreeDoc(user_doc);
			user_doc=new_user_doc;
			refresh_level_table();
			G_UNLOCK(users_doc_lock);
			time_user_file_loaded=time(NULL);
		}
	}
	return TRUE;
}

int get_access_level(const char *access_name) {
	gint* result;
	HL_LOCK_READ(levels_table_lock);
	result=g_hash_table_lookup(levels_hash_table, access_name);
	HL_UNLOCK_READ(levels_table_lock);
	return result ? *result : G_MAXINT;
}

gboolean find_user(const char *nicktofind, int *level, char **password, gboolean *unkickable, char **comment) {
	xmlXPathObjectPtr search;
	xmlNodePtr nick_node=NULL;
	GString* xpath_expr=g_string_new("");
	g_string_printf(xpath_expr,"//users/nick[@name=\"%s\"]",nicktofind);

	G_LOCK(users_doc_lock);
	search = getNodeSet (user_doc,xpath_expr->str); /* retrieve all nick */
	if (search) {
		xmlNodeSetPtr nodeset = search->nodesetval;
		if (nodeset->nodeNr != 0)
			nick_node=nodeset->nodeTab[0]; /* Only one nick must exists */
		xmlXPathFreeObject(search);
	}
	if (nick_node) {
		if (level) {
			xmlChar *xml_level=xmlGetProp(nick_node, "level");
			if (xml_level) {
				*level=atoi(xml_level);
				xmlFree(xml_level);
			}
		}
		if (password) {
			xmlChar *xml_password=xmlGetProp(nick_node, "password");
			if (xml_password) {
				*password=strdup(xml_password);
				xmlFree(xml_password);
			}
		}
		if (unkickable) {
			xmlChar *xml_unkickable=xmlGetProp(nick_node, "unkickable");
			if (xml_unkickable) {
				*unkickable=(xmlStrcmp(xml_unkickable, (const xmlChar *) "false") ? TRUE : FALSE);
				xmlFree(xml_unkickable);
			}
		}
		if (comment) {
			xmlChar *xml_comment=xmlNodeGetContent(nick_node);
			if (xml_comment) {
				*comment=strdup(xml_comment);
				xmlFree(xml_comment);
			}
		}
	}
	G_UNLOCK(users_doc_lock);
	g_string_free(xpath_expr,1);
	return nick_node ? TRUE : FALSE;
}

/*********************************************/
/* retreive the attributes of the given nick */
/*******************************************************************/
/* input: nick: is the nick to search                              */
/*        level: is a pointeur to receive the level of the user    */
/*        (can be NULL if you don't want to retreive it).          */
/*        *password: is a pointeur to receive the password of      */
/*        the user (can be NULL if you don't want to retreive it). */
/*        The password must be free()d when no more required.      */
/*        unkickable: is a pointeur to recaive a boolean that said */
/*        if the user is kickage or not (can be NULL if you don't  */
/*        want to retreive it).                                    */
/*        *comment: is a pointeur to receive comment concerning    */
/*        the user (can be NULL if you don't want to retreive it). */
/*        The comment must be free()d when no more required.       */
/*                                                                 */
/* output: TRUE if the user is found                               */
/*******************************************************************/
gboolean xml_get_user(char *nick, int *level, char **password, gboolean *unkickable, char **comment) {
	if (!refresh_user_doc())
		return FALSE;
	return find_user(nick, level, password, unkickable, comment);
}

/******************/
/* add a new user */
/********************************/
/* output: TRUE=ok, FALSE=error */
/********************************************************************/
/* note: this function doesn't perform any check of the given value */
/********************************************************************/
gboolean xml_add_user(char *nick, int level, const char *password, gboolean unkickable, const char* comment) {
	xmlNodePtr cur;
	xmlNodePtr new_user_node=NULL;
	
	if (!refresh_user_doc() || nick == NULL)
		return FALSE;
	
	G_LOCK(users_doc_lock);
	cur = xmlDocGetRootElement(user_doc)->children;
	if (cur!=NULL) {
		while (cur != NULL && (xmlStrcmp(cur->name, (const xmlChar *) "users"))) {
			cur=cur->next;
		}	
		if (cur != NULL) {
			/* I'm on the users node */
			new_user_node=xmlNewChild(cur, NULL, "nick", comment ? comment : NULL);
			xmlNewProp(new_user_node, "name", nick);
			if (level != 0) {
				xmlChar* xml_level=xmlXPathCastNumberToString(level);
				xmlNewProp(new_user_node, "level", xml_level);
				xmlFree(xml_level);
			}
			if (password != NULL)
				xmlNewProp(new_user_node, "password", password);
			if (unkickable) {
				xmlChar* xml_unkickable=xmlXPathCastBooleanToString(unkickable);
				xmlNewProp(new_user_node, "unkickable", xml_unkickable);
				xmlFree(xml_unkickable);
			}
		}
	}
	G_UNLOCK(users_doc_lock);
	return new_user_node ? TRUE : FALSE;
}

/*****************/
/* delete a user */
/********************************/
/* output: TRUE=ok, FALSE=error */
/********************************************************************/
/* note: this function doesn't perform any check of the given value */
/********************************************************************/
gboolean xml_del_user(char *nicktodel) {
	xmlXPathObjectPtr search;
	xmlNodePtr nick_node=NULL;
	
	GString* xpath_expr;

	if (!refresh_user_doc() || nicktodel == NULL)
		return FALSE;

	xpath_expr=g_string_new("");
	g_string_printf(xpath_expr,"//users/nick[@name=\"%s\"]",nicktodel);
	
	G_LOCK(users_doc_lock);
	search = getNodeSet (user_doc,xpath_expr->str); /* retrieve all nick */
	if (search) {
		xmlNodeSetPtr nodeset = search->nodesetval;
		if (nodeset->nodeNr != 0)
			nick_node=nodeset->nodeTab[0]; /* Only one nick must exists */
		xmlXPathFreeObject(search);
	}
	if (nick_node) {
		xmlUnlinkNode(nick_node);
		xmlFreeNode(nick_node);
	}
	G_UNLOCK(users_doc_lock);
	g_string_free(xpath_expr,1);
	return nick_node ? TRUE : FALSE;
}

void xml_dump_users(const char* filename) {
	G_LOCK(users_doc_lock);
	xmlSaveFormatFile("-", user_doc,1);
	G_UNLOCK(users_doc_lock);
}

void xml_free_users_resource() {
	xmlFreeDoc(user_doc);
	g_hash_table_destroy(levels_hash_table);
}
